Imports System.Drawing
Imports System.IO
Imports System.Net
Imports System.Net.Sockets
Imports System.Text

Imports UvARescue.Math
Imports System.Math



''' <summary>
''' Base class for a single agent that provides the following core 
''' functionalities:
''' - spawning the virtual robot in the game
''' - sending and recieving of messages 
''' - sensor and actor mounting
''' 
''' The Agent and all its Sensors run in the Driver's thread. Therefore
''' they assume free access to each other's members where no locking is
''' necessary. 
''' 
''' Observers attached to Agent typically run on an external thread. Therefore
''' Adding, Removing and Notifying observers IS implemented with locking.
''' 
''' </summary>
''' <remarks></remarks>
''' 
Public Class Agent

#Region " Constructor "

    Public Sub New(ByVal manifold As Manifold, ByVal name As String, ByVal agentConfig As AgentConfig, ByVal teamConfig As TeamConfig)

        If IsNothing(manifold) Then Throw New ArgumentNullException("manifold")
        If String.IsNullOrEmpty(name) Then Throw New ArgumentNullException("name")
        If IsNothing(agentConfig) Then Throw New ArgumentNullException("agentConfig")
        If IsNothing(teamConfig) Then Throw New ArgumentNullException("teamConfig")



        'by default, the agent will mount several standard sensors
        Me._StatusSensor = New StatusSensor 'no need to mount this one, uses STA messsage instead of SEN messages

        'Temp disabled to reduce load
        Me._GroundTruthSensor = New GroundTruthSensor

        'Julian RoboCup2008 Changes to USARBot.ini file mean this isn't working,
        'temporarily commented out
        Me._GPSSensor = New GPSSensor

        Me._InsSensor = New InsSensor

        Me._OdometrySensor = New OdometrySensor

        Me.Mount(Me._StatusSensor)
        Me.Mount(Me._GroundTruthSensor)

        'Temp (see note above)
        Me.Mount(Me._GPSSensor)

        Me.Mount(Me._InsSensor)
        Me.Mount(Me._OdometrySensor)


        'by default, the agent will also mount a DriveActor
        Me._DriveActor = Me.CreateDriveActor
        If Not IsNothing(Me._DriveActor) Then
            Me.Mount(Me._DriveActor)
        End If



        Me._UniqueID = Guid.NewGuid
        Me._Name = name
        Me._TeamConfig = teamConfig

        Me._AgentConfig = agentConfig

        Me._Manifold = manifold
        Me._ManifoldImage = Me.CreateManifoldImage(manifold)

        With agentConfig

            Me._Number = .AgentNumber
            Me._RobotModel = .RobotModel
            Me._Size = New Size(400, 400) 'mm
            Me._StartLocation = .StartLocation
            Me._StartRotation = .StartRotation

            'extract Pose2D from start-location and -rotation
            Dim parts() As String = Strings.Split(.StartLocation, ",")
            Dim x As Single = Single.Parse(parts(0))
            Dim y As Single = Single.Parse(parts(1))

            parts = Strings.Split(.StartRotation, ",")
            Dim rot As Single = Single.Parse(parts(2))

            Me._StartPose = New Pose2D(x, y, rot)

            'also init current pose, but beware that this should be in mm 
            SyncLock poseMutex
                Me._CurrentPoseEstimate = New Pose2D(x * 1000, y * 1000, rot)
            End SyncLock


        End With

    End Sub

    Protected Overridable Function CreateManifoldImage(ByVal manifold As Manifold) As ManifoldImage
        Return New ManifoldImage(manifold, UvARescue.Tools.Constants.MAP_RESOLUTION, False)
    End Function

#End Region

#Region " Agent Properties "

    Private _UniqueID As Guid
    Public ReadOnly Property UniqueID() As Guid
        Get
            Return _UniqueID
        End Get
    End Property

    Private _Manifold As Manifold
    Public ReadOnly Property Manifold() As Manifold
        Get
            Return Me._Manifold
        End Get
    End Property

    Private _ManifoldImage As ManifoldImage
    Public ReadOnly Property ManifoldImage() As ManifoldImage
        Get
            Return Me._ManifoldImage
        End Get
    End Property

    Private _Name As String
    Public ReadOnly Property Name() As String
        Get
            Return _Name
        End Get
    End Property

    Private _Number As Integer
    Public ReadOnly Property Number() As Integer
        Get
            Return _Number
        End Get
    End Property

    Private _RobotModel As String
    Public ReadOnly Property RobotModel() As String
        Get
            Return _RobotModel
        End Get
    End Property

    Private _Size As Size
    Public ReadOnly Property Size() As Size
        Get
            Return Me._Size
        End Get
    End Property



    Private _StartLocation As String
    Public ReadOnly Property StartLocation() As String
        Get
            Return _StartLocation
        End Get
    End Property

    Private _StartRotation As String
    Public ReadOnly Property StartRotation() As String
        Get
            Return _StartRotation
        End Get
    End Property

    Private _StartPose As Pose2D
    Public ReadOnly Property StartPose() As Pose2D
        Get
            Return Me._StartPose
        End Get
    End Property


    Private _TeamConfig As TeamConfig
    Public ReadOnly Property TeamConfig() As TeamConfig
        Get
            Return Me._TeamConfig
        End Get
    End Property

    Private _AgentConfig As AgentConfig
    Public ReadOnly Property AgentConfig() As AgentConfig
        Get
            Return Me._AgentConfig
        End Get
    End Property

#End Region

#Region " Agent Default Sensors "

    Private _StatusSensor As StatusSensor
    Protected ReadOnly Property StatusSensor() As StatusSensor
        Get
            Return Me._StatusSensor
        End Get
    End Property
    Public ReadOnly Property Status() As StatusData
        Get
            Return Me._StatusSensor.CurrentData
        End Get
    End Property

    Private _GroundTruthSensor As GroundTruthSensor
    Protected ReadOnly Property GroundTruthSensor() As GroundTruthSensor
        Get
            Return Me._GroundTruthSensor
        End Get
    End Property
    Public ReadOnly Property GroundTruth() As GroundTruthData
        Get
            Return Me._GroundTruthSensor.CurrentData
        End Get
    End Property

    Private _GPSSensor As GPSSensor
    Public ReadOnly Property GPSSensor() As GPSSensor
        Get
            Return Me._GPSSensor
        End Get
    End Property

    Private _InsSensor As InsSensor
    Protected ReadOnly Property InsSensor() As InsSensor
        Get
            Return Me._InsSensor
        End Get
    End Property
    Public ReadOnly Property Ins() As InsData
        Get
            Return Me._InsSensor.CurrentData
        End Get
    End Property

    Private _OdometrySensor As OdometrySensor
    Protected ReadOnly Property OdometrySensor() As OdometrySensor
        Get
            Return Me._OdometrySensor
        End Get
    End Property
    Public ReadOnly Property Odometry() As OdometryData
        Get
            Return Me._OdometrySensor.CurrentData
        End Get
    End Property

#End Region

#Region " Drive Actor "

    Private _DriveActor As DriveActor
    Protected ReadOnly Property DriveActor() As DriveActor
        Get
            Return Me._DriveActor
        End Get
    End Property

    Public Function CanDrive() As Boolean
        Return Not IsNothing(Me._DriveActor)
    End Function

    ''' <summary>
    ''' Override this function if you want to mount a different kind of DriveActor.
    ''' By default the StandardDriveActor will be created.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Protected Overridable Function CreateDriveActor() As DriveActor
        Return New StandardDriveActor
    End Function

    Public Sub ToggleHeadLight()
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.ToggleHeadLight()
        End If
    End Sub

    Public Sub DifferentialDrive(ByVal speed As Single, ByVal turning_speed As Single)
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.DifferentialDrive(speed, turning_speed)
        End If
    End Sub

    Public Sub Drive(ByVal speed As Single)
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.Drive(speed)
        End If
    End Sub
    Public Sub Reverse(ByVal speed As Single)
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.Drive(-speed)
        End If
    End Sub
    Public Sub Halt()
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.Drive(0)
        End If
    End Sub

    Public Sub TurnLeft(ByVal speed As Single)
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.Turn(speed)
        End If
    End Sub
    Public Sub TurnRight(ByVal speed As Single)
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.Turn(-speed)
        End If
    End Sub

    Public Sub CamUp()
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.CameraPanTilt(UvARescue.Tools.Constants.CAM_TILT_INCREMENT)
        End If
    End Sub
    Public Sub CamStop()
        Me._DriveActor.CameraPanTilt(0)
    End Sub
    Public Sub CamDown()
        If Not IsNothing(Me._DriveActor) Then
            Me._DriveActor.CameraPanTilt(-1 * UvARescue.Tools.Constants.CAM_TILT_INCREMENT)
        End If
    End Sub




#End Region

#Region " Agent Lifetime / Driver "

    Private _Driver As IDriver

    ''' <summary>
    ''' The Agent will run in the Driver's thread. Therefore the
    ''' two assume free access to each other's members. If any extnernal
    ''' thread needs access to Agent's members, thread-safety MUST
    ''' be enforced by the EXTERNAL thread.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Protected ReadOnly Property Driver() As IDriver
        Get
            Return Me._Driver
        End Get
    End Property
    Friend Sub SetDriver(ByVal driver As IDriver)
        Me._Driver = driver
    End Sub

    Public ReadOnly Property IsRunning() As Boolean
        Get
            If Not IsNothing(Me._Driver) Then
                Return Me._Driver.IsRunning
            End If
            Return False
        End Get
    End Property
    Public ReadOnly Property IsPaused() As Boolean
        Get
            If Not IsNothing(Me._Driver) Then
                Return Me._Driver.IsPaused
            End If
            Return False
        End Get
    End Property

    ''' <summary>
    ''' This is just a wrapper method. No synchronization is
    ''' enforced here, should be enforced by the Driver.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Pause()
        If Not IsNothing(Me._Driver) Then
            Me._Driver.Pause()
        End If
    End Sub

    ''' <summary>
    ''' This is just a wrapper method. No synchronization is
    ''' enforced here, should be enforced by the Driver.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub [Resume]()
        If Not IsNothing(Me._Driver) Then
            Me._Driver.Resume()
        End If
    End Sub

    Protected Friend Overridable Sub OnAgentStarted()
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifyAgentStarted()
            Next
        End SyncLock
    End Sub
    Protected Friend Overridable Sub OnAgentPaused()
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifyAgentPaused()
            Next
        End SyncLock
    End Sub
    Protected Friend Overridable Sub OnAgentResumed()
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifyAgentResumed()
            Next
        End SyncLock
    End Sub
    Protected Friend Overridable Sub OnAgentStopped()
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifyAgentStopped()
            Next
        End SyncLock
    End Sub

#End Region

#Region " Agent Sensors and Actors "

    Private sensors As New Hashtable

    Public Sub Mount(ByVal sensor As Sensor)
        Dim type As String = sensor.SensorType
        Dim table As Hashtable
        If sensors.ContainsKey(type) Then
            table = DirectCast(sensors(type), Hashtable)
        Else
            table = New Hashtable
            sensors(type) = table
        End If

        Dim name As String = sensor.SensorName
        If table.ContainsKey(name) Then
            Throw New InvalidOperationException("Sensor with this name already mounted")
        Else
            table(name) = sensor
        End If

        sensor.SetAgent(Me)
    End Sub

    Public Sub UnMount(ByVal sensor As Sensor)
        Dim type As String = sensor.SensorType
        Dim table As Hashtable
        If Not sensors.ContainsKey(type) Then
            Throw New InvalidOperationException("Sensor was not found")
        Else
            table = DirectCast(sensors(type), Hashtable)
        End If

        Dim name As String = sensor.SensorName
        If Not table.ContainsKey(name) Then
            Throw New InvalidOperationException("Sensor was not found")
        Else
            table.Remove(name)
        End If

        sensor.SetAgent(Nothing)
    End Sub

    ''' <summary>
    ''' This method is invoked by sensors when new sensor data was recieved.
    ''' Does nothing much by default, just forwards the message to observers.
    ''' Provides a hook (template method) for 
    ''' subclasses.
    ''' </summary>
    ''' <param name="sensor"></param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub NotifySensorUpdate(ByVal sensor As Sensor)
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifySensorUpdate(sensor)
            Next
        End SyncLock
        If sensor.SensorType = StatusSensor.SENSORTYPE_STATUS Then
            Me.ProcessStatusData(DirectCast(sensor, StatusSensor).CurrentData)
        ElseIf sensor.SensorType = GroundTruthSensor.SENSORTYPE_GROUNDTRUTH Then
            Me.ProcessGroundTruthData(DirectCast(sensor, GroundTruthSensor).CurrentData)
        ElseIf sensor.SensorType = GPSSensor.SENSORTYPE_GPS Then
            Me.ProcessGPSData(DirectCast(sensor, GPSSensor).CurrentData)
        End If
    End Sub

    Protected Overridable Sub ProcessStatusData(ByVal current_status As StatusData)
        If current_status.Battery < 1 AndAlso Not IsPaused Then
            Pause()
        End If
    End Sub

    Private Function NormalizeAngle(ByVal radians As Double) As Double
        While radians > PI
            radians -= 2 * PI
        End While
        While radians <= -PI
            radians += 2 * PI
        End While
        Return radians
    End Function

    Protected Overridable Sub ProcessGroundTruthData(ByVal current_data As GroundTruthData)
        Dim seedPose As New Pose2D(0, 0, 0)
        
        'acquire global ground-truth estimate
        'With Agent.GroundTruth
        seedPose.X = CType(current_data.X * 1000, Single)
        seedPose.Y = CType(current_data.Y * 1000, Single)
        seedPose.Rotation = CType(current_data.Yaw, Single)
        'End With

        'Console.WriteLine(String.Format("[Agent] - Ground Truth Position = ({0:f2} , {1:f2}) m", seedPose.X / 1000, seedPose.Y / 1000))

        'TBC Do something with Roll,Pitch
        If Abs(NormalizeAngle(current_data.Roll)) > 0.1 OrElse Abs(NormalizeAngle(current_data.Pitch)) > 0.1 Then
            Console.WriteLine(String.Format("[Agent] - Warning: tilting ({0:f2} , {1:f2}) rad", NormalizeAngle(current_data.Roll), NormalizeAngle(current_data.Pitch)))
        End If

    End Sub

    'From GPSSensor.uc
    Private ZeroZeroLocationLatDeg As Integer = 39
    Private ZeroZeroLocationLonDeg As Integer = 77 'NIST coordinates

    Private ZeroZeroLocationLatMin As Double = 8.0273
    Private ZeroZeroLocationLonMin As Double = 12.998 'NIST coordinates

    Private ScaleLatitudeDeg As Integer = 111200 ' 111 km per degree
    Private ScaleLongitudeDeg As Integer = 71670 '  72 km per degree

    Private ScaleLatitudeMin As Double = ScaleLatitudeDeg / 60
    Private ScaleLongitudeMin As Double = ScaleLongitudeDeg / 60

    Protected Overridable Sub ProcessGPSData(ByVal current_data As GPSData)
        Dim seedPose As New Pose2D(0, 0, 0)

        ' Console.WriteLine(String.Format("[Agent] - GPS received: Lateral {0:d2}:{1:f2}, Longitude {2:d2}:{3:f2}", current_data.LatDeg, current_data.LatMin, current_data.LonDeg, current_data.LonMin))

        seedPose.X = CType(current_data.LatSgn * ((current_data.LatDeg - ZeroZeroLocationLatDeg) * ScaleLatitudeDeg + (current_data.LatMin - ZeroZeroLocationLatMin) * ScaleLatitudeMin), Double)
        seedPose.Y = CType(current_data.LonSgn * ((current_data.LonDeg - ZeroZeroLocationLonDeg) * ScaleLongitudeDeg + (current_data.LonMin - ZeroZeroLocationLonMin) * ScaleLongitudeMin), Double)

        If current_data.Fix Then
            GPSSensor.X = seedPose.X
            GPSSensor.Y = seedPose.Y
            GPSSensor.Acquired = True
        Else
            GPSSensor.Acquired = False
            'Keep the current position. The lack of progress should be signaled elsewhere (SlamSeed)
        End If

        ' Console.WriteLine(String.Format("[Agent] - GPS received: distance  from ZeroZero is ({0:f2} , {1:f2}) minutes = ({2:f2} , {3:f2}) m", (current_data.LatMin - ZeroZeroLocationLatMin), (current_data.LonMin - ZeroZeroLocationLonMin), seedPose.X, seedPose.Y))


    End Sub
#End Region

#Region " Spawning and Automatic Number in Spawn Order "

    ''' <summary>
    ''' This method gets invoked by the Driver when the agent needs
    ''' to be spawned on the usarsim server.
    ''' </summary>
    ''' <remarks></remarks>
    Friend Overridable Sub Spawn()

        'construct init command
        Dim cmd As New StringBuilder()
        cmd.Append("INIT ")
        cmd.AppendFormat("{{ClassName {0}}} ", Me.GetUsarBotClassName)
        cmd.AppendFormat("{{Name {0}}} ", Me.Name)
        cmd.AppendFormat("{{Location {0}}} ", Me.StartLocation)
        cmd.AppendFormat("{{Rotation {0}}} ", Me.StartRotation)
        cmd.Append(Environment.NewLine) 'newline! very important!

        'simply use the standard interface for sending commands.
        Me.SendUsarSimCommand(cmd.ToString)

    End Sub

    Public Function GetUsarBotClassName() As String
        Return String.Format(My.Settings.UsarBotNameFormat, Me.RobotModel)
    End Function

    ''' <summary>
    ''' This member is invoked exactly once immediately after the 
    ''' usarsim server confirms that the agent was successfully spawned
    ''' in the simulator. 
    ''' 
    ''' See also the RecieveMessage routine where the NFO message is identified
    ''' as the server's confirmation message.
    ''' </summary>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub OnAgentSpawned()
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifyAgentSpawned()
            Next
        End SyncLock
    End Sub

#End Region

#Region " Send Commands to UsarSim "

    ''' <summary>
    ''' The agent should keep a queue of all its commands here. 
    ''' The driver will Dequeue the commands when sent.
    ''' </summary>
    ''' <remarks></remarks>
    Private _UsarSimCommmands As New Queue(Of String)
    Public ReadOnly Property UsarSimCommands() As Queue(Of String)
        Get
            Return Me._UsarSimCommmands
        End Get
    End Property

    ''' <summary>
    ''' The agent simply queues the commands. The driver will 
    ''' actually send the commands to the server.
    ''' </summary>
    ''' <param name="command"></param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub SendUsarSimCommand(ByVal command As String)
        If Not command.EndsWith(Environment.NewLine) Then
            command += Environment.NewLine
        End If
        Me._UsarSimCommmands.Enqueue(command)
    End Sub

#End Region

#Region " Recieve Messages "

    ''' <summary>
    ''' Invoked by the Driver whenever a new text message was recieved
    ''' from the UsarSim simulater.
    ''' </summary>
    ''' <param name="msgtype"></param>
    ''' <param name="type"></param>
    ''' <param name="name"></param>
    ''' <param name="parts"></param>
    ''' <remarks>
    ''' Actually it's a helper method in the LineParsers utility that will 
    ''' invoke this method. But conceptually it's the same.
    ''' </remarks>
    Protected Friend Overridable Sub ReceiveMessage(ByVal msgtype As MessageType, ByVal type As String, ByVal name As String, ByVal parts As Specialized.StringCollection)

        'forward the message to relevant sensors
        Select Case msgtype
            Case MessageType.Info
                'this is the very first message always recieved from usarsim
                'basically it confirms that the agent was successfully spawned
                'in the simulator
                Me.OnAgentSpawned()

            Case MessageType.Status
                Me._StatusSensor.RecieveMessage(msgtype, parts)

            Case MessageType.Unknown
                'ignore it

            Case Else
                If Not String.IsNullOrEmpty(type) AndAlso sensors.ContainsKey(type) Then
                    Dim table As Hashtable = DirectCast(sensors(type), Hashtable)
                    For Each key As String In table.Keys
                        If key = Sensor.MATCH_ANYNAME Then
                            DirectCast(table(key), Sensor).RecieveMessage(msgtype, parts)
                        ElseIf Not name = String.Empty AndAlso key = name Then
                            DirectCast(table(key), Sensor).RecieveMessage(msgtype, parts)
                        End If
                    Next
                End If


        End Select

    End Sub

    ''' <summary>
    ''' Invoked by the Driver whenever a new byte data was recieved
    ''' from the ImageServer.
    ''' </summary>
    ''' <param name="type"></param>
    ''' <param name="name"></param>
    ''' <param name="bytes"></param>
    ''' <remarks></remarks>
    Protected Friend Sub ReceiveBytes(ByVal type As String, ByVal name As String, ByVal bytes() As Byte)

        'forward the message to relevant sensors
        If Not String.IsNullOrEmpty(type) AndAlso sensors.ContainsKey(type) Then
            Dim table As Hashtable = DirectCast(sensors(type), Hashtable)
            For Each key As String In table.Keys
                If key = Sensor.MATCH_ANYNAME Then
                    DirectCast(table(key), Sensor).RecieveBytes(bytes)
                ElseIf Not name = String.Empty AndAlso key = name Then
                    DirectCast(table(key), Sensor).RecieveBytes(bytes)
                End If
            Next
        End If

    End Sub

#End Region

#Region " Manifold Sync "

    Public Sub Sync()
        Me.OnSync()
    End Sub

    Protected Overridable Sub OnSync()
    End Sub

#End Region

#Region " Observers "

    Private observers As New List(Of IAgentObserver)

    Public Sub AddObserver(ByVal observer As IAgentObserver)
        SyncLock Me.observers
            Me.observers.Add(observer)
        End SyncLock
    End Sub
    Public Sub RemoveObserver(ByVal observer As IAgentObserver)
        SyncLock Me.observers
            Me.observers.Remove(observer)
        End SyncLock
    End Sub

#End Region

#Region " Pose Estimate "

    Dim poseMutex As New Object

    Private _CurrentPoseEstimate As Pose2D
    Public ReadOnly Property CurrentPoseEstimate() As Pose2D
        Get
            SyncLock poseMutex
                Return Me._CurrentPoseEstimate
            End SyncLock
        End Get
    End Property

    ''' <summary>
    ''' Hook used by SLAM methods to notify the agent of updates pose estimates.
    ''' The agent will notify all attached observers.
    ''' </summary>
    ''' <param name="pose"></param>
    ''' <remarks></remarks>
    Public Overridable Sub NotifyPoseEstimateUpdated(ByVal pose As Pose2D)

        SyncLock poseMutex
            Me._CurrentPoseEstimate = pose
        End SyncLock

        'pose is in global coords, convert to startpose-relative coords for odometry
        Dim relative As Pose2D = pose.ToLocal(Me.StartPose)
        Me.OdometrySensor.Reset(relative.X / 1000, relative.Y / 1000, relative.Rotation)
        Me.InsSensor.Reset(pose.X / 1000, pose.Y / 1000, pose.Rotation)

        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifyPoseEstimateUpdated(pose)
            Next
        End SyncLock

    End Sub

#End Region

#Region " Alerts "

    Private _MotionTimeStamp As Date = DateTime.Now
    Public Property MotionTimeStamp() As Date
        Get
            Return Me._MotionTimeStamp
        End Get
        Set(ByVal value As Date)
            _MotionTimeStamp = value
        End Set
    End Property

    ''' <summary>
    ''' Hook used by conservativeTeleOpMotion to notify the agent of alerts.
    ''' The agent will notify all attached observers.
    ''' </summary>
    ''' <param name="alert"></param>
    ''' <remarks></remarks>
    Public Overridable Sub NotifyAlertReceived(ByVal alert As String)
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifyAlertReceived(alert)
            Next
        End SyncLock
    End Sub

#End Region

#Region " Signal Strength to Operator "

    Public Overridable Sub NotifySignalStrengthUpdated(ByVal pathloss As Double)
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifySignalStrengthReceived(pathloss)
            Next
        End SyncLock
    End Sub

#End Region

#Region " Cam Image "
    Public Overridable Sub NotifyTrackerImage(ByVal displayImage As Image)
        SyncLock Me.observers
            For Each observer As IAgentObserver In Me.observers
                observer.NotifyTrackerImage(displayImage)
            Next
        End SyncLock
    End Sub

    Public Sub CamReq()
        Me.OnCamReq()
    End Sub

    Public Sub OnCamReq()
        Dim table As Hashtable = DirectCast(sensors(CameraSensor.SENSORTYPE_CAMERA), Hashtable)
        For Each key As String In table.Keys
            If key = Sensor.MATCH_ANYNAME Then

                'DirectCast(table(key), CameraSensor).LatestCamIm.Save(String.Concat("E:\test_", otherName, ".jpg"), System.Drawing.Imaging.ImageFormat.Jpeg)
                Me.CamRepl(Me.Name, DirectCast(table(key), CameraSensor).LatestCamIm)
            End If
        Next
    End Sub

    Public Sub CamRepl(ByVal fromRobot As String, ByVal camIm As Bitmap)
        Me.OnCamRepl(fromRobot, camIm)
    End Sub

    Protected Overridable Sub OnCamRepl(ByVal fromRobot As String, ByVal camIm As Bitmap)
        'Should be Handled by CommAgent.OnCamRepl
    End Sub

    Public Overridable Sub OnCamReplReceived(ByVal otherName As String)
        Dim table As Hashtable = DirectCast(sensors(CameraSensor.SENSORTYPE_CAMERA), Hashtable)
        For Each key As String In table.Keys
            If key = Sensor.MATCH_ANYNAME Then

                'DirectCast(table(key), CameraSensor).LatestCamIm.Save(String.Concat("E:\test_", otherName, ".jpg"), System.Drawing.Imaging.ImageFormat.Jpeg)
                Me.CamRepl(otherName, DirectCast(table(key), CameraSensor).LatestCamIm)
            End If
        Next
    End Sub

#End Region

#Region " Change behavior "
    Overridable Sub ChangeBehaviour(ByVal newBehaviour As String)
    End Sub
#End Region

#Region " Finish "

    Overridable Sub Finish()
    End Sub

#End Region

End Class
