Imports UvARescue.Agent
Imports UvARescue.Math
Imports UvARescue.Tools

Imports System.Math

Public MustInherit Class IcpScanMatcher
    Implements IScanMatcher


    'for convenience 
    Public Const PI_2 As Double = PI / 2


    'break/convergence thresholds
    Protected MAX_ITERATIONS As Integer = 80
    Protected RESULTS_CONSIDERED As Integer = 10 'max num of results to consider 
    Protected RESULTS_CONVERGED As Integer = 8 'how many good results before converged
    Protected RESULTS_MAXERROR As Double = 0.01 'consider an iteration converged when rms below this


    'maximum distance within which to detect correlations
    Protected MAX_CORRELATIONDISTANCE As Single = 2000 / 4 'mm


    Protected Overridable Sub ApplyConfig(ByVal config As Tools.Config) Implements IScanMatcher.ApplyConfig
        Me.RESULTS_MAXERROR = Double.Parse(config.GetConfig("slam", "error-threshold", CStr(Me.RESULTS_MAXERROR)))
        Me.MAX_CORRELATIONDISTANCE = Single.Parse(config.GetConfig("slam", "correlation-maxdistance", CStr(Me.MAX_CORRELATIONDISTANCE)))
    End Sub



    Public Overridable Function Match(ByVal manifold As Manifold, ByVal patch1 As Patch, ByVal patch2 As Patch, ByVal seed As Math.Pose2D) As MatchResult Implements IScanMatcher.Match
        Return Me.Match(manifold, patch1.Scan, patch2.Scan, seed)
    End Function



    Public Function Match(ByVal manifold As Manifold, ByVal scan1 As ScanObservation, ByVal scan2 As ScanObservation, ByVal seed As Pose2D) As MatchResult

        If scan1.Length < 2 OrElse scan2.Length < 2 Then _
            Throw New InvalidOperationException("Cannot match scans with fewer than 2 scanpoints")

        Dim points1() As Vector2
        Dim points2() As Vector2

        If scan1.FieldOfView = scan2.FieldOfView AndAlso scan1.Resolution = scan2.Resolution Then
            'same scanner setup, reuse angles to save redundant computation
            Dim angles() As Double = Me.ComputeScanAngles(scan1)
            points1 = Me.ToPoints(scan1, angles)
            points2 = Me.ToPoints(scan2, angles)
        Else
            'differing scanner setup, compute angles for each scan individually
            points1 = Me.ToPoints(scan1, Me.ComputeScanAngles(scan1))
            points2 = Me.ToPoints(scan2, Me.ComputeScanAngles(scan2))
        End If


        Dim filter1() As Boolean = Me.ToFilter(scan1)
        Dim filter2() As Boolean = Me.ToFilter(scan2)
        Dim dangle As Double = scan1.FieldOfView / scan1.Length

        Return Me.MatchPoints(manifold, points1, points2, seed, dangle, filter1, filter2)

    End Function


    Protected Friend MustOverride Function MatchPoints(ByVal manifold As Manifold, ByVal points1() As Vector2, ByVal points2() As Vector2, ByVal seed As Pose2D, ByVal dangle As Double, ByVal filter1() As Boolean, ByVal filter2() As Boolean) As MatchResult


    Protected Overridable Function HasConverged(ByVal numPoints As Integer, ByVal numCorrespondencies As Integer, ByVal numIterations As Integer) As Boolean
        Return numIterations < MAX_ITERATIONS
    End Function



    Protected Overridable Function ToPoints(ByVal scan As ScanObservation, ByVal angles() As Double) As Vector2()
        With scan
            Dim points(.Range.Length - 1) As Vector2
            'Dim xoffset As Double = scan.OffsetX * .Factor
            'Dim yoffset As Double = scan.OffsetY * .Factor

            Dim dist As Double, angle As Double
            For i As Integer = 0 To .Range.Length - 1
                dist = .Range(i) * .Factor
                angle = angles(i)
                points(i) = New Vector2(Cos(angle) * dist, Sin(angle) * dist)
            Next

            Return points
        End With
    End Function

    Protected Overridable Function ToFilter(ByVal scan As ScanObservation) As Boolean()
        With scan
            Dim filter(.Range.Length - 1) As Boolean
            For i As Integer = 0 To .Range.Length - 1
                filter(i) = .Range(i) < .MinRange OrElse .Range(i) > .MaxRange
            Next
            Return filter
        End With
    End Function

    Protected Overridable Function ToLocal(ByVal points() As Vector2, ByVal target As Pose2D) As Vector2()
        Dim rotmx As TMatrix2D = target.ToGlobalMatrix

        Dim local(points.Length - 1) As Vector2
        For i As Integer = 0 To points.Length - 1
            local(i) = rotmx * points(i)
        Next
        Return local

    End Function



    Protected Overridable Function ComputeScanAngles(ByVal scan As ScanObservation) As Double()

        Dim fov As Double = scan.FieldOfView
        Dim res As Double = scan.Resolution

        Dim length As Integer = CInt(fov / res) + 1
        Dim fromAngle As Double = -fov / 2

        Dim angles(length - 1) As Double
        For i As Integer = 0 To length - 1
            angles(i) = fromAngle + res * i
        Next
        Return angles

    End Function





    Protected Function CorrelatePointsDM(ByVal manifold As Manifold, ByVal points1() As Vector2, ByVal points2() As Vector2, ByVal filter1() As Boolean, ByVal filter2() As Boolean) As Correlation(Of Vector2)()

        Dim maxdist As Double = Pow(MAX_CORRELATIONDISTANCE, 2)

        'dists will hold the full distance matrix
        Dim dists(points1.Length - 1, points2.Length - 1) As Double

        '- collect preferences for points1
        '- and simultaneously construct full distance matrix dists
        Dim marks1to2(points1.Length - 1) As Integer
        For i As Integer = 0 To points1.Length - 1

            Dim curdist As Double = maxdist + 1
            Dim curmark As Integer = -1

            For j As Integer = 0 To points2.Length - 1

                'compute distance and store in full distance matrix
                Dim dist As Double = Me.ComputeSquaredDistance(points1(i), points2(j))
                dists(i, j) = dist

                'check preference
                If Not filter1(i) AndAlso Not filter2(j) _
                    AndAlso dist < maxdist AndAlso dist < curdist Then

                    curdist = dists(i, j)
                    curmark = j

                End If

            Next

            'store preference
            marks1to2(i) = curmark

        Next

        'matches will get to store the number of pairs that prefer each other
        Dim matches As Integer = 0

        Dim correlations As New List(Of Correlation(Of Vector2))

        'collect preferences for points2
        Dim marks2to1(points2.Length - 1) As Integer
        For j As Integer = 0 To points2.Length - 1

            Dim curdist As Double = maxdist + 1
            Dim curmark As Integer = -1

            'find nearest point of those that prefer me
            For i As Integer = 0 To points1.Length - 1
                If marks1to2(i) = j Then
                    Dim dist As Double = dists(i, j)
                    If dist < maxdist AndAlso dist < curdist Then
                        curdist = dist
                        curmark = i
                    End If
                End If
            Next

            marks2to1(j) = curmark

            If curdist < maxdist Then
                'we have a match!
                correlations.Add(New Correlation(Of Vector2)(points1(curmark), points2(j)))
                matches += 1
            End If

        Next

        'collect all pairs
        Dim pairs(matches - 1, 1) As Integer
        Dim p As Integer = 0
        For j As Integer = 0 To points2.Length - 1

            If marks2to1(j) < 0 Then
                Continue For
            End If

            pairs(p, 0) = marks2to1(j)
            pairs(p, 1) = j

            p += 1

        Next

        Return correlations.ToArray

    End Function

    Protected Function CorrelatePointsQT(ByVal manifold As Agent.Manifold, ByVal points1() As Math.Vector2, ByVal points2() As Math.Vector2, ByVal filter1() As Boolean, ByVal filter2() As Boolean) As Correlation(Of Math.Vector2)()

        Dim qtree1 As New QuadTree(Of Vector2)
        For i As Integer = 0 To points1.Length - 1
            If Not filter1(i) Then
                qtree1.Insert(points1(i))
            End If
        Next

        Dim qtree2 As New QuadTree(Of Vector2)
        For i As Integer = 0 To points2.Length - 1
            If Not filter2(i) Then
                qtree2.Insert(points2(i))
            End If
        Next

        Dim correlations As New List(Of Correlation(Of Vector2))

        Dim point1 As Vector2, point2 As Vector2
        For Each point As Vector2 In points2

            point1 = qtree1.FindNearestNeighbour(point, MAX_CORRELATIONDISTANCE)
            If IsNothing(point1) Then Continue For

            point2 = qtree2.FindNearestNeighbour(point1, MAX_CORRELATIONDISTANCE)
            If IsNothing(point2) Then Continue For

            If point2 Is point Then
                'we found a pair
                correlations.Add(New Correlation(Of Vector2)(point1, point2))
            End If
        Next

        Return correlations.ToArray

    End Function




    Protected Overridable Function ComputeSquaredDistance(ByVal point1 As Vector2, ByVal point2 As Vector2) As Double
        Return Pow(point2.X - point1.X, 2) + Pow(point2.Y - point1.Y, 2)
    End Function

End Class
