Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.IO
Imports System.Windows.Forms

Imports UvARescue.Agent
Imports UvARescue.Tools

Public Class SaveMapDialog

    Private _GeoRef As GeoReference
    Private _Image As ManifoldImage
    Private _ManifoldView As ManifoldView
    Private _TeamMembers As List(Of String)
    Private _PatchIndex As Dictionary(Of String, List(Of Patch))
    Private _PatchesToBeRendered As Dictionary(Of String, Integer)
    Private _SaveMapPath As String

#Region " Constructor "
    Public Sub New(ByVal teamconfig As TeamConfig, ByRef view As ManifoldView)
        ' This call is required by the Windows Form Designer.
        InitializeComponent()

        ' Initialize Tabs
        With tabMenu
            .TabPages.Item(0).Text = "Process"
            .TabPages.Item(1).Text = "Save"
        End With

        ' Save image locally for processing
        Me._Image = view.Image()

        ' Save ManifoldView locally.  NOTE: passed by reference so changes are permanent!!
        Me._ManifoldView = view

        ' Initialize list of team members (excluding Operator)
        Me.ResetTeammembers(teamconfig)

        ' Initialize Processing Tab
        Me.InitializeProcessTab(view.Image.Manifold.PatchIndex)

        ' Initialize Saving Tab
        Me.InitializeSavingTab()
    End Sub
#End Region

#Region " Constructor helper functions "
    Private Sub InitializeSavingTab()
        'try to load last used pathname
        Me._SaveMapPath = My.Settings.LastMapSavePath
        If Not My.Computer.FileSystem.DirectoryExists(Me._SaveMapPath) Then
            Me._SaveMapPath = My.Computer.FileSystem.Drives(0).ToString
        End If
        Me.lblPath.Text = Me._SaveMapPath

        'check for Geo Reference Data
        If Me._Image.GetGeoReferenceData(Me._GeoRef) Then
            With Me._GeoRef
                Me.txtOffsetX.Text = CStr(.OffsetX)
                Me.txtOffsetY.Text = CStr(.OffsetY)
                Me.txtPixelWidth.Text = CStr(.ScaleX)
                Me.txtPixelHeight.Text = CStr(.ScaleY)
                Me.txtShearingX.Text = CStr(.ShearX)
                Me.txtShearingY.Text = CStr(.ShearY)
            End With

        Else
            MsgBox("Could not load GeoReference data for Saving from ManifoldImage", MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
            Me.DialogResult = Windows.Forms.DialogResult.Cancel
            Me.Close()

        End If
    End Sub

    Private Sub InitializeProcessTab(ByVal patchIndex As Dictionary(Of String, List(Of Patch)))
        ' Save the Patch index locally
        Me._PatchIndex = patchIndex

        ' Initialize dictionary of how many patches should be rendered for each agent
        Me._PatchesToBeRendered = New Dictionary(Of String, Integer)

        ' Initialize labels and patches to be rendered;
        ' Add Patch Track Bar for each agent
        For i As Integer = 0 To Me._TeamMembers.Count - 1
            Dim currName As String = Me._TeamMembers(i)
            Dim currPatchTotal As Integer
            If (patchIndex.ContainsKey(currName)) Then
                currPatchTotal = patchIndex.Item(currName).Count
            Else
                currPatchTotal = 0
            End If

            Me.SetPatchTotal(currName, currPatchTotal)

            With Me.pnlAgentPatches.Controls
                Dim patchView As AgentPatchController = Me.CreateAgentPatchView(currName, currPatchTotal)
                patchView.Dock = DockStyle.Top
                .Add(patchView)
                .SetChildIndex(patchView, 0)
            End With
        Next
    End Sub

    Public Sub ResetTeammembers(ByVal teamconfig As TeamConfig)
        Me._TeamMembers = New List(Of String)
        If Not IsNothing(teamconfig) Then
            Dim members() As String = Strings.Split(teamconfig.TeamMembers, ",")
            For Each member As String In members
                Dim parts() As String = Strings.Split(member, "-")
                Dim memberName As String = parts(1)
                If Not (memberName = "Operator") Then
                    Me._TeamMembers.Add(memberName)
                End If
            Next
        End If
    End Sub
#End Region

#Region " Save image "
    Private Sub Save()
        'get all supported save options
        Dim optvalues As Array = System.Enum.GetValues(GetType(ManifoldImage.SaveOption))
        Dim optnames() As String = System.Enum.GetNames(GetType(ManifoldImage.SaveOption))

        'inspect selected image formats
        Dim formats As New List(Of ImageFormat)
        For Each item As ListViewItem In Me.lvwFormats.Items
            If item.Checked Then
                formats.Add(Me.GetImageFormat(CStr(item.Tag)))
            End If
        Next

        'to display progress, count the number of tasks
        Dim tasks As Integer = formats.Count
        tasks *= optvalues.Length 'each option for every format
        tasks *= 2 'world file generation for each format+save option
        If Me.chkGeoTiff.Checked Then
            tasks += 1 'GeoTIFF generation
        End If

        'setup progressbar
        With Me.barProgress
            .Value = 0
            .Minimum = 0
            .Maximum = tasks
        End With

        'construct base filename
        Dim basefilename As String = Me.txtName.Text.Trim
        If String.IsNullOrEmpty(basefilename) Then
            MsgBox("No filename specified", MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
            Exit Sub
        Else
            'combine with selected folder
            basefilename = Path.Combine(Me._SaveMapPath, Path.GetFileNameWithoutExtension(basefilename))
        End If

        'now save, loop through formats
        For Each format As ImageFormat In formats

            'for each save option
            For i As Integer = 0 To optvalues.Length - 1
                'get save option
                Dim optvalue As ManifoldImage.SaveOption = CType(optvalues.GetValue(i), ManifoldImage.SaveOption)

                'construct filenames for this layer
                Dim imgFilename As String = String.Format("{0}-{1}-[{2}].{3}", basefilename, CInt(optvalue), optnames(i).ToLower, Me.GetImageFileExtension(format))
                Dim wldFilename As String = String.Format("{0}-{1}-[{2}].{3}", basefilename, CInt(optvalue), optnames(i).ToLower, Me.GetWorldFileExtension(format))

                'save the world file
                Me._GeoRef.SaveWorldFile(wldFilename)
                Me.barProgress.Value += 1

                Dim bitmap As Bitmap = Nothing
                If Me._Image.GetImageFileSaveData(bitmap, optvalue) Then
                    Try
                        bitmap.Save(imgFilename, format)
                    Catch ex As Exception
                        MsgBox("Could not Save as " & format.ToString.ToUpper & " to " & imgFilename & vbNewLine & ex.Message, MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
                    Finally
                        If Not IsNothing(bitmap) Then
                            bitmap.Dispose()
                        End If
                    End Try
                Else
                    MsgBox("Could not get data from Manifold for layer " & optnames(i), MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
                End If
                Me.barProgress.Value += 1
            Next
        Next

        If Me.chkGeoTiff.Checked Then
            Dim bitmap As Bitmap = Nothing
            If Me._Image.GetImageFileSaveData(bitmap, ManifoldImage.SaveOption.Jury) Then
                Try
                    GeoTiff.Save(basefilename & Me.GetImageFileExtension(ImageFormat.Tiff), bitmap, Me._GeoRef)
                Catch ex As Exception
                    MsgBox("Could not Save as GeoTIFF" & vbNewLine & ex.Message, MsgBoxStyle.OkOnly, MsgBoxStyle.Exclamation)
                End Try
            Else
                MsgBox("Could not get data from Manifold for GeoTIFF", MsgBoxStyle.OkOnly, MsgBoxStyle.Exclamation)
            End If
            Me.barProgress.Value += 1
        End If

        MsgBox("Map was saved successfully", MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
        Me.barProgress.Value = 0
    End Sub
#End Region

#Region " Image saving helper functions "

    Private Function GetImageFormat(ByVal extension As String) As ImageFormat
        Select Case extension.ToLower
            Case "tif", "tiff"
                Return ImageFormat.Tiff
            Case "gif"
                Return ImageFormat.Gif
            Case "jpg", "jpe", "jpeg"
                Return ImageFormat.Jpeg
            Case "png"
                Return ImageFormat.Png
            Case Else
                Return ImageFormat.Bmp
        End Select
    End Function

    Private Function GetImageFileExtension(ByVal format As ImageFormat) As String
        If format.Equals(ImageFormat.Tiff) Then
            Return "tif"
        ElseIf format.Equals(ImageFormat.Gif) Then
            Return "gif"
        ElseIf format.Equals(ImageFormat.Jpeg) Then
            Return "jpg"
        ElseIf format.Equals(ImageFormat.Png) Then
            Return "png"
        Else
            Return "bmp"
        End If
    End Function

    Private Function GetWorldFileExtension(ByVal format As ImageFormat) As String
        Dim imExt As String = Me.GetImageFileExtension(format)
        Dim wfExt As String = ""
        If imExt.Length > 2 Then
            'first char of imgext + last char of imgext + w 
            wfExt &= imExt.Substring(0, 1)
            wfExt &= imExt.Substring(imExt.Length - 1, 1)
            wfExt &= "w"
        Else
            'default to TFW
            wfExt = "tfw"
        End If
        Return wfExt
    End Function
#End Region

#Region " Map processing helper functions "
    Public Sub SetPatchTotal(ByVal name As String, ByVal patchTotal As Integer)
        If Not (Me._PatchesToBeRendered.ContainsKey(name)) Then
            Me._PatchesToBeRendered.Add(name, patchTotal)
        Else
            Me._PatchesToBeRendered.Item(name) = patchTotal
        End If
    End Sub

    Protected Overridable Function CreateAgentPatchView(ByVal name As String, ByVal patchTotal As Integer) As AgentPatchController
        Dim view As New AgentPatchController(Me, name, patchTotal)
        Return view
    End Function
#End Region

#Region " Actions "

    Private Sub btnSave_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnSave.Click
        Try
            Me.Save()
        Catch ex As Exception
            MsgBox("Error occurred during Save" & vbNewLine & ex.Message, MsgBoxStyle.OkOnly, MsgBoxStyle.Exclamation)
        End Try
    End Sub

    Private Sub btnOK_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnOK.Click
        Me.DialogResult = Windows.Forms.DialogResult.Abort
        Me.Close()
    End Sub

    Private Sub btnPath_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnPath.Click
        With Me.dlgFolderBrowser
            Dim folder As String = My.Settings.LastMapSavePath
            If Not String.IsNullOrEmpty(folder) Then
                .SelectedPath = folder
            End If
            Select Case .ShowDialog
                Case Windows.Forms.DialogResult.OK
                    Me.lblPath.Text = .SelectedPath
                    Me._SaveMapPath = .SelectedPath
                    My.Settings.LastMapSavePath = .SelectedPath
                    My.Settings.Save()
                Case Else
                    Exit Sub
            End Select
        End With
    End Sub

#End Region

#Region " Redraw Map "
    Public Sub RedrawMap()
        Dim tempManifold As Manifold = New Manifold()
        Dim tempManifoldImage As ManifoldImage = New ManifoldImage(tempManifold, UvARescue.Tools.Constants.MAP_RESOLUTION, False)
        Dim currPatch As Patch

        For Each currAgent As String In Me._TeamMembers
            For i As Integer = 0 To Me._PatchesToBeRendered.Item(currAgent) - 1
                currPatch = Me._PatchIndex.Item(currAgent).Item(i)
                tempManifold.Extend(currPatch)
                tempManifoldImage.RenderPatch(currPatch)
            Next
        Next

        Me._Image = tempManifoldImage
        Me._ManifoldView.Image = tempManifoldImage
        Me._ManifoldView.Invalidate()
    End Sub
#End Region

End Class