Imports System.Drawing
Imports System.IO
Imports System.Windows.Forms

Imports UvARescue.Agent
Imports UvARescue.Slam
Imports UvARescue.Tools
Imports UvARescue.Math
Imports UvARescue.ImageAnalysis
Imports UvARescue.Agent.BehaviorControl

Public Class AgentController
    Inherits UserControl
    Implements IAgentObserver

#Region " Constructor / Destructor "

    Public Sub New(ByVal teamConfig As TeamConfig, ByVal usarOperator As UsarOperatorAgent, ByVal agentName As String, ByVal manifold As Manifold)

        ' This call is required by the Windows Form Designer.
        InitializeComponent()

        ' Add any initialization after the InitializeComponent() call.
        Me.SetStyle(ControlStyles.AllPaintingInWmPaint, True)
        Me.SetStyle(ControlStyles.OptimizedDoubleBuffer, True)
        Me.SetStyle(ControlStyles.UserPaint, True)

        Me._TeamConfig = teamConfig
        Me._AgentName = agentName
        Me._AgentConfig = New UsarAgentConfig(agentName)

        'try to load last used config file
        Dim cfgfile As String = My.Settings.LastAgentConfigFile
        If File.Exists(cfgfile) Then
            Try
                Me._AgentConfig.Load(cfgfile)
            Catch ex As Exception
            End Try
        End If



        Me._Manifold = manifold
        Me._UsarOperator = usarOperator
        Me._RunStatus = UsarRunStatus.Preparing

        Me.LoadBehaviors()
        If Not TypeOf Me._Agent Is UsarOperatorAgent AndAlso Not Me._AgentConfig.BehaviorMode Is Nothing Then
            Me.ShowCurrentBehavior(Me._AgentConfig.BehaviorMode)
        End If

        Me.ResetDrivePanel()
        Me.ResetLabels()
        Me.ResetSpeedLabel()

    End Sub

    'UserControl overrides dispose to clean up the component list.
    <System.Diagnostics.DebuggerNonUserCode()> _
    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        Try
            If disposing AndAlso components IsNot Nothing Then
                components.Dispose()
            End If
        Finally
            MyBase.Dispose(disposing)
        End Try
    End Sub

#End Region

#Region " Team Configuration "

    Private _TeamConfig As TeamConfig
    Public ReadOnly Property TeamConfig() As TeamConfig
        Get
            Return Me._TeamConfig
        End Get
    End Property

#End Region

#Region " Agent Configuration "

    Private _AgentName As String
    Public Property AgentName() As String
        Get
            Return Me._AgentName
        End Get
        Set(ByVal value As String)
            If Not value = Me._AgentName Then
                Me._AgentName = value
                Me._AgentConfig.AgentName = value
                Me.ResetLabels()
            End If
        End Set
    End Property

    Private _AgentConfig As UsarAgentConfig
    Public ReadOnly Property AgentConfig() As UsarAgentConfig
        Get
            Return Me._AgentConfig
        End Get
    End Property

    Private Sub ShowAgentConfigView()
        Dim dialog As New AgentConfigDialog(Me._AgentConfig)
        Dim result As DialogResult = dialog.ShowDialog(Me)
        Select Case result
            Case DialogResult.OK
                Me.ResetLabels()
        End Select
    End Sub

#End Region

#Region " Monitor Current RunStatus "

    Private _RunStatus As UsarRunStatus
    Protected ReadOnly Property RunStatus() As UsarRunStatus
        Get
            Return Me._RunStatus
        End Get
    End Property

    Public Sub NotifyRunStatusChanged(ByVal status As UsarRunStatus)
        Me.OnRunStatusChanged(status)
    End Sub

    Protected Overridable Sub OnRunStatusChanged(ByVal status As UsarRunStatus)
        Me._RunStatus = status

        Me.ResetActionButton()
        Me.ResetLabels()

        Select Case status
            Case UsarRunStatus.Reporting, UsarRunStatus.Done
                Me.StopAgent()
        End Select

    End Sub

    Private Sub ResetActionButton()
        With Me.btnAction
            Select Case Me._RunStatus
                Case UsarRunStatus.Preparing, UsarRunStatus.Ready
                    .Enabled = (Me._RunStatus = UsarRunStatus.Preparing)
                    .Image = My.Resources.Gear

                Case UsarRunStatus.Running
                    .Enabled = True
                    If IsNothing(Me._Agent) Then
                        .Image = My.Resources.Lightning
                    Else
                        If Me._Agent.IsPaused Then
                            .Image = My.Resources.Play
                        Else
                            .Image = My.Resources.Pause

                            If Me._AgentConfig.AgentName = "Operator" Then
                                Me.btnDrive.Visible = False
                            Else
                                Me.btnDrive.Visible = True
                            End If
                        End If
                    End If


                Case UsarRunStatus.Reporting, UsarRunStatus.Done
                            .Enabled = False
                            .Image = My.Resources.Play

            End Select
        End With
    End Sub

    Private Sub btnAction_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnAction.Click
        Select Case Me._RunStatus
            Case UsarRunStatus.Preparing
                Me.ShowAgentConfigView()

            Case UsarRunStatus.Ready
                'no associated action

            Case UsarRunStatus.Running
                If IsNothing(Me._Agent) Then
                    Me.StartAgent()
                Else
                    If Me.Agent.IsPaused Then
                        Me.Agent.Resume()
                    Else
                        Me.Agent.Pause()
                    End If
                End If

            Case UsarRunStatus.Reporting, UsarRunStatus.Done

        End Select
    End Sub

#End Region

#Region " Monitoring Agent Status "

    Public Sub NotifyAgentStarted(ByVal agent As Agent.Agent)
        If TypeOf agent Is UsarOperatorAgent Then
            If Not agent Is Me._UsarOperator Then

                'new operator agent was spawned
                Dim [operator] As UsarOperatorAgent = DirectCast(agent, UsarOperatorAgent)

                If Not IsNothing(Me._Agent) AndAlso TypeOf Me._Agent Is UsarProxyAgent Then
                    'register proxy (if necessary)
                    Dim proxy As UsarProxyAgent = DirectCast(Me._Agent, UsarProxyAgent)
                    If Not IsNothing(Me._UsarOperator) Then
                        'unregister from current operator
                        Me._UsarOperator.UnRegisterProxy(proxy)
                    End If
                    If Not IsNothing([operator]) Then
                        'register with new operator
                        [operator].RegisterProxy(proxy)
                    End If

                End If

                'update local reference
                Me._UsarOperator = [operator]

            End If
        End If
    End Sub

    Public Sub NotifyAgentStopped(ByVal agent As Agent.Agent)
        If TypeOf agent Is UsarOperatorAgent Then
            If agent Is Me._UsarOperator Then

                'operator agent was stopped
                Dim [operator] As UsarOperatorAgent = DirectCast(agent, UsarOperatorAgent)

                If Not IsNothing(Me._Agent) AndAlso TypeOf Me._Agent Is UsarProxyAgent Then
                    'unregister proxy 
                    [operator].UnRegisterProxy(DirectCast(Me._Agent, UsarProxyAgent))
                End If

                'update local reference
                Me._UsarOperator = [operator]

            End If
        End If
    End Sub

    Private _Manifold As Manifold
    Public ReadOnly Property Manifold() As Manifold
        Get
            Return Me._Manifold
        End Get
    End Property

    Private _UsarOperator As UsarOperatorAgent
    Public ReadOnly Property UsarOperator() As UsarOperatorAgent
        Get
            Return Me._UsarOperator
        End Get
    End Property

    Private _Driver As IDriver
    Public ReadOnly Property Driver() As IDriver
        Get
            Return Me._Driver
        End Get
    End Property

    Private _Agent As Agent.Agent
    Public ReadOnly Property Agent() As Agent.Agent
        Get
            Return Me._Agent
        End Get
    End Property


    Public Event AgentStarted As EventHandler(Of EventArgs(Of Agent.Agent))
    Public Event AgentStopped As EventHandler(Of EventArgs(Of Agent.Agent))


    Private Sub StartAgent()
        If Not IsNothing(Me._Agent) Then Throw New InvalidOperationException("Agent already initialized")

        With Me._AgentConfig

            Me._Agent = .CreateAgent(Me._Manifold, Me._TeamConfig, AgentStartMode.FromCommanderGUI)
            Me._Agent.AddObserver(Me)

            If TypeOf Me._Agent Is UsarOperatorAgent Then
                Me._UsarOperator = DirectCast(Me._Agent, UsarOperatorAgent)
            ElseIf TypeOf Me._Agent Is UsarProxyAgent Then
                If Not IsNothing(Me._UsarOperator) Then
                    Me._UsarOperator.RegisterProxy(DirectCast(Me._Agent, UsarProxyAgent))
                End If
            End If

            Me.ShowCurrentBehavior(.BehaviorMode)

            Me._Driver = .CreateDriver(Agent, Me._TeamConfig)

        End With

        Me._Driver.Start()

        Me.ResetDrivePanel()

        RaiseEvent AgentStarted(Me, New EventArgs(Of Agent.Agent)(Me._Agent))

    End Sub

    Private Sub StopAgent()
        If Not IsNothing(Me._Driver) AndAlso Me._Driver.IsRunning Then
            Me._Driver.Stop()
            If Not IsNothing(Me._Agent) Then
                Me._Agent.RemoveObserver(Me)

                If TypeOf Me._Agent Is UsarProxyAgent Then
                    If Not IsNothing(Me._UsarOperator) Then
                        Me._UsarOperator.UnRegisterProxy(DirectCast(Me._Agent, UsarProxyAgent))
                    End If
                End If


                RaiseEvent AgentStopped(Me, New EventArgs(Of Agent.Agent)(Me._Agent))

                Me._Agent = Nothing
            End If

            Me._Driver = Nothing
            Me.ResetDrivePanel()

        End If
    End Sub


    Private _Pathloss As String = ""

    Private Sub ResetLabels()

        Me.lblTitle.Text = String.Format("#{2}: {0} - [{1}]", Me._AgentName, Me._AgentConfig.RobotModel, Me._AgentConfig.AgentNumber)

        Select Case Me._RunStatus
            Case UsarRunStatus.Preparing
                'abuse various labels to display current config
                With Me._AgentConfig
                    Me.lblPose.Text = String.Format("{0} m", .StartLocation)
                    Me.lblInfo.Text = String.Format("{0} rad", .StartRotation)
                    Me.lblInfo.ForeColor = Color.White
                    If .LogPlayback Then
                        Me.lblPower.Text = Path.GetFileName(.LogFile)
                    Else
                        Dim opts As String = ""
                        If .UseImageServer Then opts += " +ImgServer"
                        Me.lblPower.Text = "Live" & opts
                    End If
                    Me.lblLight.Text = String.Format("{0} +{1} +{2}", .MappingMode, .ScanMatcher, .SeedMode)
                    Me.ShowCurrentBehavior(.BehaviorMode)
                End With

            Case UsarRunStatus.Ready
                'configuration done
                Me.lblInfo.Text = "Configured"
                Me.lblInfo.ForeColor = Color.Yellow
                Me.ShowCurrentBehavior(Me._AgentConfig.BehaviorMode)

            Case UsarRunStatus.Reporting, UsarRunStatus.Done
                Me.lblInfo.Text = "Stopped"
                Me.lblInfo.ForeColor = Color.Red

            Case Else

                Dim info As String
                Dim fclr As Color
                If IsNothing(Me._Agent) Then
                    info = "Ready to Spawn"
                    fclr = Color.Yellow
                Else
                    If Not Me._Agent.IsRunning Then
                        info = "Stopped"
                        fclr = Color.Red
                    Else
                        If Me._Agent.IsPaused Then
                            info = "Paused"
                            fclr = Color.Yellow
                        Else
                            info = "Running"
                            fclr = Color.LawnGreen
                        End If
                    End If
                End If

                Me.lblInfo.Text = info & Me._Pathloss
                Me.lblInfo.ForeColor = fclr

        End Select
    End Sub

    Private CamImageMutex As New Object

    Private Delegate Sub SensorUpdateHandler(ByVal sensor As Sensor)
    Public Sub NotifySensorUpdate(ByVal sensor As Sensor) Implements IAgentObserver.NotifySensorUpdate
        Try
            'make sure we are on the UI thread
            If Me.InvokeRequired Then
                If sensor.SensorType = CameraSensor.SENSORTYPE_CAMERA Then
                    Dim handler As New SensorUpdateHandler(AddressOf Me.NotifySensorUpdate)
                    Me.BeginInvoke(handler, sensor)
                End If

            Else

                SyncLock Me.CamImageMutex
                    If Me._ShowCameraImages Then
                        If sensor.SensorType = CameraSensor.SENSORTYPE_CAMERA Then

                            Dim image As Image = DirectCast(sensor, CameraSensor).LatestCamIm
                            If Not image Is Nothing Then
                                'Me.pnlCamImage.BackgroundImage = image
                            End If
                            'Me.pnlCamImage.Invalidate()

                        End If
                    End If
                End SyncLock

            End If
        Catch ex As Exception
            Debug.WriteLine(ex.ToString)

        End Try

    End Sub

    Private Delegate Sub PoseUpdateHandler(ByVal pose As Math.Pose2D)
    Public Sub NotifyPoseEstimateUpdated(ByVal pose As Math.Pose2D) Implements IAgentObserver.NotifyPoseEstimateUpdated

        'make sure we are on the UI thread
        If Me.InvokeRequired Then
            Dim handler As New PoseUpdateHandler(AddressOf Me.NotifyPoseEstimateUpdated)
            Me.BeginInvoke(handler, pose)

        Else

            Dim estimate As String = pose.ToString(2)
            If Not IsNothing(Me._Agent) Then
                Dim truth As String = String.Format("{0:f2} , {1:f2}", Me._Agent.GroundTruth.X, Me._Agent.GroundTruth.Y)
                Me.lblPose.Text = String.Format("{0} ({1}) m", estimate, truth)

                Me.lblPower.Text = String.Format("{0} secs", Me._Agent.Status.Battery)
                Me.lblLight.Text = DirectCast(IIf(Me._Agent.Status.LightToggle, "On", "Off"), String)
            End If

        End If
    End Sub

    Private Delegate Sub AgentAlertHandler(ByVal alert As String)
    Public Sub NotifyAlertReceived(ByVal alert As String) Implements IAgentObserver.NotifyAlertReceived

        'make sure we are on the UI thread
        If Me.InvokeRequired Then
            Dim handler As New AgentAlertHandler(AddressOf Me.NotifyAlertReceived)
            Me.BeginInvoke(handler, alert)

        Else
            Me.lblAlert.Text = "Alert: " + alert
            Me.lblAlert.Visible = True

        End If
    End Sub

    Private Delegate Sub AgentStatusChangedHandler()
    Public Sub NotifyAgentStatusChanged() Implements IAgentObserver.NotifyAgentStarted, IAgentObserver.NotifyAgentSpawned, IAgentObserver.NotifyAgentPaused, IAgentObserver.NotifyAgentResumed, IAgentObserver.NotifyAgentStopped

        'make sure we are on the UI thread
        If Me.InvokeRequired Then
            Dim handler As New AgentStatusChangedHandler(AddressOf Me.NotifyAgentStatusChanged)
            Me.BeginInvoke(handler)

        Else

            Me.ResetActionButton()
            Me.ResetLabels()

        End If
    End Sub

    Private Delegate Sub StrengthReceivedHandler(ByVal pathloss As Double)
    Public Sub NotifySignalStrengthReceived(ByVal pathloss As Double) Implements IAgentObserver.NotifySignalStrengthReceived

        'make sure we are on the UI thread
        If Me.InvokeRequired Then
            Dim handler As New StrengthReceivedHandler(AddressOf Me.NotifySignalStrengthReceived)
            Me.BeginInvoke(handler, pathloss)

        Else

            Me._Pathloss = String.Format(" {0} dBm", pathloss)
            Me.ResetLabels()

        End If

    End Sub

#End Region

#Region " Changing Behavior "
    Private Sub LoadBehaviors()
        Dim behaviors As Array = System.Enum.GetNames(GetType(BehaviorType))
        Dim behavior As String
        For Each behavior In behaviors
            Me.BehaviorDisplay.Items.Add(behavior)
        Next
    End Sub

    Private Sub ShowCurrentBehavior(ByVal currBehavior As String)
        If Me._AgentConfig.AgentName = "Operator" Then
            Me.BehaviorDisplay.Visible = False
        Else 'TypeOf Me._Agent Is UsarProxyAgent Or TypeOf Me._Agent Is UsarAgent Then
            Me.BehaviorDisplay.Visible = True
        End If
        Me.BehaviorDisplay.SelectedItem = currBehavior
    End Sub

    Private Sub BehaviorDisplay_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles BehaviorDisplay.SelectedIndexChanged
        Dim newBehavior As String
        If Me.BehaviorDisplay.SelectedItem Is Nothing Then
            newBehavior = "TeleOperation"
        Else
            newBehavior = Me.BehaviorDisplay.SelectedItem.ToString
        End If
        If Not (newBehavior = Me._AgentConfig.BehaviorMode) Then
            Me._AgentConfig.BehaviorMode = newBehavior
            If Not IsNothing(Me._Agent) Then
                Me._UsarOperator.CommActor.RelayBehaviorChangeCommand(Me._UsarOperator.Name, Me._Agent.Name, newBehavior)
            End If
            Me.ShowCurrentBehavior(newBehavior)
        End If
    End Sub
#End Region

#Region " Driving "

    Private Sub ResetDrivePanel()
        Dim enable As Boolean = True
        enable = enable AndAlso Not IsNothing(Me._Agent)
        enable = enable AndAlso Me._Agent.IsRunning

        If enable Then
            Me.btnDrive.Enabled = True
        Else
            Me.btnDrive.Enabled = False
        End If

    End Sub

    Private Sub btnDrive_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnDrive.Click
        Me.SuspendLayout()
        Me.pnlControl.Hide()
        Me.pnlDrive.Show()
        Me.ResumeLayout(True)
    End Sub

    Private Sub btnCancelDrive_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnCancelDrive.Click
        Me.SuspendLayout()
        Me.pnlDrive.Hide()
        Me.pnlControl.Show()
        Me.ResumeLayout(True)
    End Sub

    Private Sub btnHeadLight_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnHeadLight.Click
        Me.Agent.ToggleHeadLight()
    End Sub

    Private _Speed As Single = 2.0F

    Private Sub ResetSpeedLabel()
        Me.lblSpeed.Text = String.Format("{0:f2} rad/s", Me._Speed)
    End Sub

    Private Sub btnSpeedUp_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnSpeedUp.Click
        Me._Speed += 0.25F
        Me.ResetSpeedLabel()
    End Sub
    Private Sub btnSpeedDown_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnSpeedDown.Click
        Me._Speed -= 0.25F
        Me.ResetSpeedLabel()
    End Sub


    Private Sub btnForward_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnForward.Click
        If Not Me.Agent Is Nothing Then
            Me.Agent.Drive(Me._Speed)
        End If
    End Sub
    Private Sub btnReverse_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnReverse.Click
        If Not Me.Agent Is Nothing Then
            Me.Agent.Reverse(1.5)
        End If
    End Sub
    Private Sub btnStop_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnStop.Click
        If Not Me.Agent Is Nothing Then
            Me.Agent.Halt()
        End If
    End Sub

    Private Sub btnTurnLeft_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnTurnLeft.Click
        If Not Me.Agent Is Nothing Then
            Me.Agent.TurnLeft(0.5)
        End If
    End Sub
    Private Sub btnTurnRight_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnTurnRight.Click
        If Not Me.Agent Is Nothing Then
            Me.Agent.TurnRight(0.5)
        End If
    End Sub

    Private Sub btnSync_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnSync.Click
        If Not Me.Agent Is Nothing Then
            Me.Agent.Sync()
        End If
    End Sub

    Private Sub alertAcknowledged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnForward.Click, btnReverse.Click, btnTurnLeft.Click, btnTurnRight.Click
        Me.lblAlert.Visible = False
        Me.Agent.MotionTimeStamp = DateTime.Now
    End Sub

#End Region

#Region "Tracker Button"
    Private Sub Button1_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button1.Click
        'Activate tracker boolean in the usarFollowAgent
        'Get USARFollowAgent object, and call function
        If TypeOf Me._Agent Is UsarSkinDetAgent Then
            DirectCast(Me._Agent, UsarSkinDetAgent).activateTracker()
        End If
    End Sub
#End Region

#Region "Tracker drawing functions"
    Private Delegate Sub trackUpdateHandler(ByVal displayImage As Image)
    Public Sub NotifyTrackerImage(ByVal displayImage As Image) Implements IAgentObserver.NotifyTrackerImage
        If Me.InvokeRequired Then
            Dim handler As New trackUpdateHandler(AddressOf Me.NotifyTrackerImage)
            Me.BeginInvoke(handler, displayImage)
        Else
        End If

    End Sub
#End Region

#Region " Camera Control Functionality "

    Private _ShowCameraImages As Boolean = True

    Private Sub btnCamToggle_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnCamToggle.Click
        If (Me._ShowCameraImages) Then
            Me._ShowCameraImages = False
            Me.btnCamToggle.Image = My.Resources.CameraOff
        Else
            Me._ShowCameraImages = True
            Me.btnCamToggle.Image = My.Resources.CameraOn
        End If
    End Sub

    Private _camImgIndex As Integer = 1
    Private Sub btnCamReq_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnCamReq.Click
        'If Not Me.Agent Is Nothing Then
        ' Me.Agent.CamReq()
        'End If
        If Not Me.pnlCamImage.BackgroundImage Is Nothing Then
            Dim camIm As System.Drawing.Image = Me.pnlCamImage.BackgroundImage
            Dim currPose As Pose2D = Me.Agent.CurrentPoseEstimate
            Dim filename As String = "D:\USARSim\Results\View" + Me._camImgIndex.ToString + "_" + CInt(currPose.X / 100).ToString + "_" + CInt(currPose.Y / 100).ToString + "_" + CInt(currPose.Rotation * 180 / System.Math.PI).ToString + ".jpg"
            camIm.Save(filename, Drawing.Imaging.ImageFormat.Jpeg)
            Me._camImgIndex += 1
        End If
    End Sub

    Private Sub btnCamUp_MouseDown(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles btnCamUp.MouseDown
        If Not Me.Agent Is Nothing Then
            Me.Agent.CamUp()
        End If
    End Sub

    Private Sub btnCamDown_MouseDown(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles btnCamDown.MouseDown
        If Not Me.Agent Is Nothing Then
            Me.Agent.CamDown()
        End If
    End Sub

    '    Private Sub btnCamStopMovement(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles btnCamUp.MouseUp, btnCamDown.MouseUp
    '        If Not Me.Agent Is Nothing Then
    '            Me.Agent.CamStop()
    '        End If
    '    End Sub

#End Region

End Class
